from messagebus import MessageBus
from order_api import OrderService
from typing import Optional

message_bus = None  # type: Optional[MessageBus]


def create_order_service(
    pos_id,     # type: int
):
    return OrderService(
        message_bus=message_bus,
        pos_id=pos_id,
    )
