from setuptools import (
    setup,
    find_packages,
)


setup(
    name="NReportsApi",
    version="1.2.0",
    packages=find_packages(),
    install_requires=[],
    url="https://www.e-deploy.com.br",
    author="Lucas de Lima",
    author_email="lucas.lima@e-deploy.com.br"
)
