class ErrorBuildingDataBuilder(Exception):

    def __init__(
        self,
        report_name: str,
        error_message: str,
    ) -> None:
        message = f"Error building data builder for report {report_name}: {error_message}"
        super(ErrorBuildingDataBuilder, self).__init__(message)
