class ContextKeys(object):
    is_in_contingency = "is_in_contingecy"
    contingency_datetime = "contingency_datetime"
    contingency_reason = "contingency_reason"
    total_icms_vbc = "total_icms_vbc"
    total_prod = "total_prod"
    total_icms = "total_icms"
    total_pis = "total_pis"
    total_cofins = "total_cofins"
    item_no = "item_no"
    current_sale_item = "current_sale_item"
    sale_items = "sale_items"
    fiscal_number = "fiscal_number"
    emission_type = "emission_type"
    nfce_key = "nfce_key"
    data_emissao = "data_emissao"
    v_nf = "v_nf"
    digest_value = "digest_value"
    cpf_cnpj = "cpf_cnpj"
    random_number = "random_number"
    dv = "dv"
    csrt = "csrt"
    id_csrt = "id_csrt"
    deson = "deson"
    total_fcp = "total_fcp"
    total_fcp_st = "total_fcp_st"
    total_fcp_ant = "total_fcp_ant"
    delivery_partner = "delivery_partner"
