class ReportFileNotFound(Exception):

    def __init__(
        self,
        report_name: str,
    ) -> None:
        super(ReportFileNotFound, self).__init__(f"No JSON found for requested report: {report_name}")
