from abc import (
    ABC,
    abstractmethod,
)

from ._EventPublishDetail import EventPublishDetail


class EventProcessorListener(ABC):

    @abstractmethod
    def event_processed(self, publish_detail: EventPublishDetail) -> None:
        raise NotImplementedError()
